﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Net.Sockets;

namespace MockASocket
{
	/// <summary>
	/// A class that can be used to connect and log in to an FTP server using a Socket connection.
	/// </summary>
	class FTPConnection : IDisposable
	{
		/// <summary>
		/// Possible FTP return codes.
		/// </summary>
		public enum ReturnCodes
		{
			/// <summary>
			/// Data connection already open; transfer starting.
			/// </summary>
			DataConnectionOpen = 125,

			/// <summary>
			/// File status okay; about to open data connection.
			/// </summary>
			FileStatusOK = 150,

			/// <summary>
			/// Command okay.
			/// </summary>
			CommandOK = 200,

			/// <summary>
			/// Command not implemented, superfluous at this site
			/// </summary>
			CommandNotImplemented = 202,

			/// <summary>
			/// File status.
			/// </summary>
			FileStatus = 213,

			/// <summary>
			/// Service ready for new user.
			/// </summary>
			ServiceReady = 220,

			/// <summary>
			/// Closing data connection. Requested file action successful (for example, file transfer or file abort).
			/// </summary>
			ClosingDataConnection = 226,

			/// <summary>
			/// Entering Passive Mode (h1,h2,h3,h4,p1,p2).
			/// </summary>
			EnteringPassiveMode = 227,

			/// <summary>
			/// User logged in, proceed. Logged out if appropriate.
			/// </summary>
			UserLoggedIn = 230,

			/// <summary>
			/// Requested file action okay, completed.
			/// </summary>
			RequestedFileActionOK = 250,

			/// <summary>
			/// "PATHNAME" created.
			/// </summary>
			PathCreated = 257,

			/// <summary>
			/// User name okay, need password.
			/// </summary>
			UsernameOK = 331,

			/// <summary>
			/// Requested file action pending further information
			/// </summary>
			RequestedFileActionPending = 350
		}

		/// <summary>
		/// The FTP port number.
		/// </summary>
		public const int FTP_PORT = 21;

		/// <summary>
		/// The size of the buffers for the Sockets.
		/// </summary>
		public const int BLOCK_SIZE = 256;

		private ISocket _socket;
		private string _host;
		private string _username;
		private string _password;

		/// <summary>
		/// Default constructor using System.Net.Sockets.Socket via the SocketAdapter.
		/// </summary>
		/// <param name="host">The FTP host.</param>
		public FTPConnection(string host, string username, string password) : this(new SocketAdapter(), host, username, password)
		{ }

		/// <summary>
		/// Constructor using custom (e.g. mocked) Socket.
		/// </summary>
		/// <param name="socket">The Socket to use.</param>
		/// <param name="host">The FTP host.</param>
		public FTPConnection(ISocket socket, string host, string username, string password)
		{
			_socket = socket;
			_host = host;
			_username = username;
			_password = password;
			OpenSocket();
		}

		/// <summary>
		/// Opens the socket connection to the FTP server.
		/// </summary>
		private void OpenSocket()
		{
			_socket.Connect(_host, FTP_PORT);
			string[] response = ReadFromSocket();
			if (GetReturnCodeFromResponse(response) != ReturnCodes.ServiceReady)
			{
				throw new Exception("Error while connecting to FTP server. Server responded: " + String.Join(" ", response));
			}
		}

		/// <summary>
		/// Reads the response from the FTP server from the Socket.
		/// </summary>
		/// <returns>The server's response message.</returns>
		private string[] ReadFromSocket()
		{
			MemoryStream ms = new MemoryStream();
			byte[] buffer = new byte[BLOCK_SIZE];
			while (true)
			{
				int bytes = _socket.Receive(buffer, buffer.Length, SocketFlags.None);
				ms.Write(buffer, 0, bytes);
				if (bytes < buffer.Length)
				{
					break;
				}
			}

			StreamReader sr = new StreamReader(ms);
			ms.Position = 0;
			string response = sr.ReadToEnd();
			sr.Close();
			ms.Close();
			string[] result = response.Split(new char[] { '\n' });
			Console.WriteLine("Response from Server: " + result[0]);
			return result;
		}

		/// <summary>
		/// Parses the server's response for the return code.
		/// </summary>
		/// <param name="response">The response message.</param>
		/// <returns>The return code.</returns>
		private ReturnCodes GetReturnCodeFromResponse(string[] response)
		{
			try
			{
				return (ReturnCodes)Enum.Parse(typeof(ReturnCodes), response[0].Substring(0, 3));
			}
			catch (Exception e)
			{
				throw new Exception("Error while parsing response from FTP server: " + e.Message);
			}
		}

		/// <summary>
		/// Sends an (FTP) command through the socket.
		/// </summary>
		/// <param name="command">The command.</param>
		/// <returns>The return code from the server.</returns>
		private ReturnCodes SendCommand(String command)
		{
			Byte[] cmdBytes = Encoding.ASCII.GetBytes((command + "\r\n").ToCharArray());
			_socket.Send(cmdBytes, cmdBytes.Length, SocketFlags.None);
			return GetReturnCodeFromResponse(ReadFromSocket());
		}

		/// <summary>
		/// Logs in to the FTP server.
		/// </summary>
		public void Login()
		{
			ReturnCodes rc = SendCommand("USER " + _username);
			if (rc != ReturnCodes.UsernameOK)
			{
				throw new Exception("Could not log in with username '" + _username + "'.");
			}
			rc = SendCommand("PASS " + _password);
			if (rc != ReturnCodes.UserLoggedIn)
			{
				throw new Exception("Wrong password for username '" + _username + "'.");
			}
		}

		/// <summary>
		/// Closes the connection to the FTP server.
		/// </summary>
		public void Close()
		{
			SendCommand("QUIT");
		}

		#region IDisposable Member

		public void Dispose()
		{
			if (_socket != null)
			{
				_socket.Close();
			}
		}

		#endregion
	}
}
