﻿using NUnit.Framework;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Net.Sockets;
using Rhino.Mocks;

namespace MockASocket
{
	[TestFixture]
	class FTPConnectionTest
	{
		private const string FTP_HOST = "fake.ftp.com";
		private const string FTP_USER = "fakeuser";
		private const string FTP_PASS = "fakepass";

		MockRepository _mocks;
		ISocket _mockSocket;

		/// <summary>
		/// The SetUp method that is called before each test.
		/// </summary>
		[SetUp]
		public void SetUp()
		{
			_mocks = new MockRepository();
			_mockSocket = _mocks.StrictMock<ISocket>();
			_FTPMessages = new List<String>();
		}

		/// <summary>
		/// Tests the opening of a new FTP connection.
		/// </summary>
		[Test]
		public void OpenConnection()
		{
			ExpectConnect(_mockSocket);

			ReplayAll();

			FTPConnection ftp = new FTPConnection(_mockSocket, FTP_HOST, FTP_USER, FTP_PASS);
			_mocks.VerifyAll();
		}

		/// <summary>
		/// Tests logging in to the FTP server.
		/// </summary>
		[Test]
		public void Login()
		{
			ExpectConnect(_mockSocket);

			ExpectSend(_mockSocket, "USER " + FTP_USER);
			ExpectReceive(_mockSocket, "331 Enter server password for fake server");

			ExpectSend(_mockSocket, "PASS " + FTP_PASS);
			ExpectReceive(_mockSocket, "230 Connected to fake server");

			ReplayAll();

			FTPConnection ftp = new FTPConnection(_mockSocket, FTP_HOST, FTP_USER, FTP_PASS);
			ftp.Login();
			_mocks.VerifyAll();
		}

		/// <summary>
		/// Tests closing the FTP connection.
		/// </summary>
		[Test]
		public void Close()
		{
			ExpectConnect(_mockSocket);

			ExpectSend(_mockSocket, "QUIT");
			ExpectReceive(_mockSocket, "226 Closing fake connection");

			ReplayAll();

			FTPConnection ftp = new FTPConnection(_mockSocket, FTP_HOST, FTP_USER, FTP_PASS);
			ftp.Close();
			_mocks.VerifyAll();
		}

		#region Delegate for the Receive method (needed because the parameter "buffer" is modified)
		private delegate int ReceiveDelegate(byte[] buffer, int size, SocketFlags flags);
		/// <summary>
		/// A delegate for Receive that is called by ExpectReceive.
		/// Copies the response message to the buffer and returns the size.
		/// </summary>
		/// <param name="buffer"></param>
		/// <param name="size"></param>
		/// <param name="flags"></param>
		/// <returns></returns>
		private int Receive(byte[] buffer, int size, SocketFlags flags)
		{
			_FTPMessageIterator.MoveNext();
			byte[] str = Encoding.ASCII.GetBytes(_FTPMessageIterator.Current);
			for (int i = 0; i < str.Length; i++)
				buffer[i] = str[i];
			return str.Length;
		}
		#endregion

		#region Needed to pass messages to the Receive delegate
		private IList<String> _FTPMessages;
		private IEnumerator<String> _FTPMessageIterator;
		#endregion

		#region Helper methods (to avoid duplicated code)
		private void ReplayAll()
		{
			_mocks.ReplayAll();
			_FTPMessageIterator = _FTPMessages.GetEnumerator();
		}

		/// <summary>
		/// Expect a call of the Connect method and simulate a correct response by the FTP server.
		/// </summary>
		/// <param name="mockSocket">The fake socket.</param>
		private void ExpectConnect(ISocket mockSocket)
		{
			mockSocket.Connect(FTP_HOST, FTPConnection.FTP_PORT);
			ExpectReceive(mockSocket, "220 This is a fake FTP server");
		}

		/// <summary>
		/// Expect a call of the Receive method and simulate a response by the FTP server using the given message.
		/// </summary>
		/// <param name="mockSocket">The fake socket.</param>
		/// <param name="message">The message to return.</param>
		private void ExpectReceive(ISocket mockSocket, string message)
		{
			_FTPMessages.Add(message);
			Expect.Call(mockSocket.Receive(new byte[FTPConnection.BLOCK_SIZE], FTPConnection.BLOCK_SIZE, 0)).Do(new ReceiveDelegate(Receive));
		}

		/// <summary>
		/// Expect a call of the Send method with the given message.
		/// </summary>
		/// <param name="mockSocket">The fake socket.</param>
		/// <param name="message">The expected message.</param>
		private void ExpectSend(ISocket mockSocket, string message)
		{
			byte[] buffer = new byte[FTPConnection.BLOCK_SIZE];
			buffer = Encoding.ASCII.GetBytes(message + "\r\n");
			Expect.Call(mockSocket.Send(buffer, buffer.Length, 0)).Return(buffer.Length);
		}
		#endregion
	}
}
